#Copyright STACKEO - INRIA 2020 .
from abc import abstractmethod

from Stkml.Stkml import Stkml


class RegionsMap:

    @abstractmethod
    def from_stackml(self, stackml: Stkml, output: str):
        pass



import os
from zipfile import ZipFile, ZipInfo

class PackageZip(ZipFile):

    def extract_all(self, path=None, members=None, pwd=None):
        if members is None:
            members = self.namelist()

        if path is None:
            path = os.getcwd()
        else:
            path = os.fspath(path)

        for member in members:
            zinfo = self.getinfo(member)
            ret_val = self._extract_member(zinfo, path, pwd)
            attr = zinfo.external_attr >> 16
            os.chmod(ret_val, attr)

    def zip(self, path: str) -> None:
        ziph = self
        for root, _, files in os.walk(path):
            for file in files:
                file_name = os.path.join(root, file)
                with open(file_name, 'rb') as file_pointer:
                    info = ZipInfo(file_name.replace(f'{path}/', ""))
                    info.external_attr = 0o500 << 16
                    ziph.writestr(info, file_pointer.read())
        ziph.close()



#Copyright STACKEO - INRIA 2020 .
import json
import os

import yaml
from jinja2 import Environment
from jsonschema import validate

from Stkml import STKML_METADATA, STKML_PACKAGES


class StkmlPackage:

    def __init__(self, package: str):
        self.package_dir = package
        self.name: str = ''
        self.version: str = ''
        self.author: str = ''
        self.get_name_from_path()

    def get_name_from_path(self):
        self.name = os.path.basename(self.package_dir)

    def generate_package(self, environment: Environment) -> None:
        package_template = environment.get_template('StkmlIn/package.stkml.yaml.j2')
        package = package_template.render()
        package_file = os.path.join(self.package_dir, f'{self.name}.stkml.yaml')
        with open(package_file, 'w') as file:
            file.write(package)

    def generate_metadata(self, environment: Environment) -> None:
        metadata_template = environment.get_template('StkmlIn/stkml.yaml.j2')
        metadata = metadata_template.render(package=self)
        metadata_file = os.path.join(self.package_dir, 'stkml.yaml')
        with open(metadata_file, 'w') as file:
            file.write(metadata)

    @classmethod
    def check_metadata(cls, metadata_path):
        if not os.path.isfile(metadata_path):
            raise FileNotFoundError(f'can not find stkml metadata in {metadata_path}')
        with open(metadata_path, 'r') as meta_file:
            metadata = yaml.load(meta_file, Loader=yaml.FullLoader)
            with open(STKML_METADATA, 'r') as file:
                metadata_schema = json.load(file)
                validate(metadata, metadata_schema)
    @classmethod
    def get_packages_dir(cls, stkml_folder):
        stkml_folder = os.path.abspath(stkml_folder)
        packages_dir = os.path.join(stkml_folder, STKML_PACKAGES)
        if not os.path.isdir(packages_dir):
            stkml_folder = os.path.join(stkml_folder, os.pardir)
            try:
                return cls.get_packages_dir(stkml_folder)
            except RecursionError as error:
                raise RecursionError('can not find .stkml_packages folder ') from error
        return os.path.realpath(packages_dir)



#Copyright STACKEO - INRIA 2020 .
# pylint: disable=too-many-public-methods,too-many-nested-blocks, too-many-locals)
import os
from typing import List
import copy
import json

import requests
from PIL import Image
from jinja2 import Environment, Template


import jsonschema
from jsonschema import validate
import yaml
from yaml.scanner import ScannerError

from Stkml import STKML_URL_SCHEMA, STKML_EXTENSION, ICONS
from Stkml.Stkml.LayerElement import Component
from Stkml.Stkml.ModelException import ModelException

from Stkml.Stkml.Region import Region
from Stkml.Stkml.StkmlPackage import StkmlPackage
from Stkml.Stkml.StkmlSyntaxicError import StkmlSyntaxicError
from Stkml.Stkml.StkmlSyntaxicErrorList import StkmlSyntaxicErrorList
from Stkml.Stkml.Tier import Tier
from Stkml.Stkml.Link import Link
from Stkml.Stkml.Node import Node
from Stkml.Stkml.Topology import Topology
from Stkml.Stkml.WarningList import WarningList


class Stkml:

    def __init__(self, folder=None):
        self.stkmlfile = None
        if folder:
            self.stkmlfile = 'main'
            self.stkml_folder = folder
        self.tiers: List[Tier] = []
        self.links: List[Link] = []
        self.regions: List[Region] = []
        self.topology = Topology()
        self.schema = self.get_stkml_schema()
        self.errors = []
        self.libraries = {}

    def check_stkml(self, file) -> bool or ScannerError:
        if not file['ignore-check']:
            if file['file']:
                self.check_file(file['path'])
            else:
                for root, _, _file in os.walk(file['path']):
                    for file_ in _file:
                        self.check_file(os.path.join(root, file_))
        return True

    def check_file(self, file, check_other_files=False) -> None or ScannerError or IOError:
        if '.stkml.yaml' in file:
            with open(file, 'r') as stkml_file:
                yaml.load(stkml_file, Loader=yaml.FullLoader)
        elif os.path.basename(file) == 'stkml.yaml':
            pass
        else: #image
            if check_other_files:
                try:
                    image = Image.open(file)
                except IOError as error:
                    raise IOError(f'{file} is not accepted format ') from error
                else:
                    image.close()

    def check_stkml_package_file(self, file) -> bool or ScannerError or IOError:
        if not file['ignore-check']:
            if file['file']:
                self.check_file(file['path'], True)
            else:
                for root, _, _file in os.walk(file['path']):
                    for file_ in _file:
                        self.check_file(os.path.join(root, file_), True)
        return True

    def get_library(self, name):#-> FileNotFoundError or str
        library = {'path': '', 'file': True, 'ignore-check': False}
        local_file = os.path.join(self.stkml_folder, f'{name}{STKML_EXTENSION}')
        if os.path.isfile(local_file):
            library['path'] = local_file
        else:
            local_folder = os.path.join(self.stkml_folder, name)
            if os.path.isdir(local_folder):
                library['path'] = local_folder
                library['file'] = False
            else:
                try:
                    stkml_packages = StkmlPackage.get_packages_dir(self.stkml_folder)
                except RecursionError as error :
                    raise FileNotFoundError(
                        f'The {name} package does not exists') from error
                global_file = os.path.join(stkml_packages, f'{name}{STKML_EXTENSION}')
                if os.path.isfile(global_file):
                    self.libraries[name] = global_file
                    library['path'] = global_file
                    library['ignore-check'] = True
                else:
                    global_folder = os.path.join(stkml_packages, name)
                    if os.path.isdir(global_folder):
                        library['path'] = global_folder
                        library['file'] = False
                        library['ignore-check'] = True
                    else:
                        raise FileNotFoundError(
                            f'The {name} package does not exists')
        return library

    def get_stkml_schema(self) -> dict or None:
        try:
            schema = None
            request = requests.get(STKML_URL_SCHEMA)
        except requests.exceptions.ConnectionError:
            pass
        else:
            if request.status_code == 200:
                schema = json.loads(request.content.decode("utf-8"))
            request.close()
            return schema
    def validate_stkml(self, file) -> bool or Exception:
        if not file['ignore-check']:
            if file['file']:
                self._validate(file['path'])
            else:
                for root, _, _file in os.walk(file['path']):
                    for file_ in _file:
                        if '.stkml.yaml' in file_:
                            self._validate(os.path.join(root, file_))

    def validate_package(self, file) -> Exception or list:
        warns = {}
        if not file['ignore-check']:
            if file['file']:
                path = file['path']
                self._validate_package(path, warns)
            else:
                for root, _, _file in os.walk(file['path']):
                    for file_ in _file:
                        if '.stkml.yaml' in file_:
                            path = os.path.join(root, file_)
                            self._validate_package(path, warns)
        return warns
    def _validate(self, stkml_file):
        with open(stkml_file, 'r') as stkml_:
            stkml = yaml.load(stkml_, Loader=yaml.FullLoader)
            try:
                validate(stkml, self.schema)
            except jsonschema.exceptions.ValidationError:
                self.errors.append(StkmlSyntaxicError(self.schema, stkml, stkml_file))

    def _validate_package(self, stkml_file, warnings):
        with open(stkml_file, 'r') as stkml_:
            stkml = yaml.load(stkml_, Loader=yaml.FullLoader)
            try:
                validate(stkml, self.schema)
            except jsonschema.exceptions.ValidationError:
                self.errors.append(StkmlSyntaxicError(self.schema, stkml, stkml_file))
            else:
                if stkml:
                    model_def = stkml.get('modeldef')
                    if model_def:
                        node_models = model_def.get('NodeModels')
                        if node_models:
                            warns = []
                            for node in node_models:
                                icon = node.get('icon')
                                _, warn = self.get_node_path_icon(icon, None, stkml_file)
                                if warn:
                                    msg = f'{icon} for {node.get("name")} node'
                                    warns.append(msg)
                    if warns:
                        warnings[os.path.basename(stkml_file)] = warns


    def check_stkml_project(self) -> str or ScannerError:
        lib = self.get_library(self.stkmlfile)
        self.libraries[self.stkmlfile] = lib
        self.check_stkml(lib)
        i = 0
        with open(lib['path'], 'r') as stkml_file:
            stkml = yaml.load(stkml_file, Loader=yaml.FullLoader)
            for import_ in (stkml.get('import') or []):
                lib = self.get_library(import_)
                self.libraries[import_] = lib
                self.check_stkml(lib)
                i += 1
        return True

    def check_stkml_package(self, check_metadata=False) -> str or ScannerError:
        lib = self.get_library(self.stkml_folder)
        self.check_stkml_package_file(lib)
        self.libraries[self.stkml_folder] = lib
        if check_metadata:
            StkmlPackage.check_metadata(os.path.join(self.stkml_folder, 'stkml.yaml'))
        return True

    def validate_stkml_package(self) -> bool or list or StkmlSyntaxicErrorList:
        if self.schema:
            warnings = {}
            for lib in self.libraries:
                warn = self.validate_package(self.libraries[lib])
                if warn:
                    warnings.update(warn)
            if len(self.errors) > 0:
                raise StkmlSyntaxicErrorList(self.errors)
            if len(warnings) > 0:
                return warnings
            return True
        raise Exception('Stkml schema does not found please verify your connection')

    def validate_stkml_project(self) -> str or StkmlSyntaxicErrorList:
        if self.schema:
            for lib in self.libraries:
                self.validate_stkml(self.libraries[lib])

            if len(self.errors) > 0:
                raise StkmlSyntaxicErrorList(self.errors)
            return True
        raise Exception('Stkml schema does not found please verify your connection')


    def add_tier(self, tier_type: str) -> Tier:
        for tier in self.tiers:
            if tier.type == tier_type:
                return tier
        tier = Tier(name=tier_type, type_=tier_type)
        self.tiers.append(tier)
        return tier

    def create_architecture(self):
        with open(self.libraries.get(self.stkmlfile)["path"], 'r') as stkml_file:
            stkml = yaml.load(stkml_file, Loader=yaml.FullLoader)
            topology = stkml['topology']
            self.topology.name = topology.get('name')
            self.topology.usecase = topology.get('useCase')
            self.topology.type = topology.get('type')
            nodes = topology.get('Nodes')
            warns = []
            if nodes:
                warns = self.create_tiers(nodes)
            links = topology.get('Links')
            if links:
                self.create_links(links)
            regions = topology.get('Regions')
            if regions:
                self.create_regions(regions)
            return warns


    def create_regions(self, regions) -> None:
        for region in regions:
            region_ = Region(name=region.get('name'), type_=region.get('type'))
            for node in region.get('NodePopulation'):
                region_.add_node(node_id=node.get('nodeId'), number=node.get('number'))
            self.regions.append(region_)

    def create_links(self, links) -> None:
        for link in links:
            source = link.get('source')
            sink = link.get('sink')
            self.links.append(Link(source=source, sink=sink,
                                   layer=link.get('layer'), secure=link.get('secure')))
    def get_node_model(self, model_name):
        node = None
        if '.' in model_name:#extrenel model

            model = model_name.split('.')
            file = self.get_model_path(model[0])
            if file:
                with open(file, 'r') as s_file:
                    model_def = yaml.load(s_file, Loader=yaml.FullLoader)
                node_id = model[1]
            else:
                raise ModelException(msg=f'{model[0]} can not be imported')
        else:#internal model
            file = self.libraries.get(self.stkmlfile)['path']
            with open(file, 'r') as s_file:
                model_def = yaml.load(s_file, Loader=yaml.FullLoader)
            node_id = model_name
        model_def = model_def.get('modeldef')
        if model_def:
            model_def = model_def.get('NodeModels')
            if model_def:
                for node_def in model_def:
                    if node_def.get('name') == node_id:
                        node = node_def
                        break
            else:
                raise ModelException(msg=f'The are no NodeModels on {file}')
        else:
            raise ModelException(msg=f'The are no modeldef for {node_id}')
        if not node:
            raise ModelException(msg=f'The {node_id} model does not exist on {file}')
        return node, file

    def get_model_path(self, model):
        model_splited = model.split('/')
        model_path = None
        for i in range(len(model_splited)):
            path = "/".join(model_splited[:i+1])
            model_path = self.libraries.get(path)
            if model_path:
                if model_path['file']:
                    model_path = model_path['path']
                else:
                    model_path = f"{model_path['path']}/{'/'.join(model_splited[i+1:])}{STKML_EXTENSION}"
                    if not os.path.isfile(model_path):
                        model_path = None
                    break

        return model_path

    def create_tiers(self, nodes) -> None:
        warnings = []
        for node in nodes:
            model = node.get('model')
            node_def, file_path = self.get_node_model(model)
            node_type = node_def.get('technology')
            tier_type = node.get('tier')
            #if tier not in node_type:
            #    warnings.append(f"you are using {node_type} in {tier} Tier")
            tier = self.add_tier(tier_type)
            node_ = tier.add_node(id_=node.get('id'), type_=node_type,
                                  cardinality=node.get('cardinality'))
            node_.set_nature(node_def.get('nature'))
            icon, warning = self.get_node_path_icon(icon=node_def.get('icon'),
                                           tier_type=tier_type, file_path=file_path)
            if warning:
                warnings.append(warning)
            node_.set_icon(icon)
            for layer in node_.element_types:
                elem = node_def.get(layer)
                if elem:
                    hub = elem.get('ElementType')
                    if not hub:
                        hub = 'EndPoint'
                    layer_elem = node_.add_layer_element(layer, type_=layer, hub=hub)
                    layer_elem.set_components(elem.get('Components'))
        self.sort_tiers()
        return warnings

    def get_node_path_icon(self, icon: str, tier_type: str, file_path: str):
        icon_path = os.path.join(ICONS, f'{tier_type}.png')
        warning = None
        if icon:
            icon = os.path.join(os.path.dirname(file_path), icon)
            if not os.path.isfile(icon):
                warning = FileNotFoundError(icon)
            else:
                icon_path = icon
        return icon_path, warning

    def generate_files(self, environment: Environment) -> Template:
        template = environment.get_template('StkmlIn/main.stkml.yaml.j2')
        return template.render(Stackml=self)

    def get_node(self, id_: str) -> Node:
        node = None
        for tier in self.tiers:
            for node_ in tier.nodes:
                if node_.id_ == id_:
                    node = node_
                    break
        return node


    def get_node_tier(self, id_: str) -> Tier or None:
        tier = None
        for tier_ in self.tiers:
            for node in tier_.nodes:
                if node.id_ == id_:
                    tier = tier_
                    break
        return tier

    def get_tier(self, tier_name: str) -> Tier or None:
        tier = None
        for tier_ in self.tiers:
            if tier_.name == tier_name:
                tier = tier_
                break
        return tier

    def get_component(self, layer, cmp_id) -> Component or None:
        comp = None
        for tier in self.tiers:
            for node in tier.nodes:
                cmps = node.get_components(layer)
                for component in (cmps or []):
                    if component.id_ == cmp_id:
                        comp = component
                        break
                else:
                    continue
                break
        return comp

    def creat_referenced_components(self) -> None:
        for i_t in range(len(self.tiers)):
            tier = self.tiers[i_t]
            for i_n in range(len(tier.nodes)):
                node = tier.nodes[i_n]
                for i_layer in range(len(node.layers_element)):
                    layer = node.layers_element[i_layer]
                    for i_comp in range(len(layer.components)):
                        comp = layer.components[i_comp]
                        id_ = comp.id_ref
                        if id_:

                            comp = self.get_component(layer.type, id_)
                            id_ = comp.id_
                            comp = copy.deepcopy(comp)
                            layer.components[i_comp] = comp
                            layer.components[i_comp].id_ = id_


    def from_drawio(self, drawio) -> None:
        for tier in drawio.tiers:
            stackml_tier = Tier(name=tier.name, type_=tier.type)
            self.tiers.append(stackml_tier)
            for node in tier.nodes:
                stackml_node = stackml_tier.add_node(id_=node.name, cardinality=1, type_=node.type)
                for layer in reversed(node.layers):
                    if len(layer.components) > 0:
                        stackml_layer = stackml_node.add_layer_element(element='', type_=layer.type, hub=layer.hub)
                        for component in layer.components:
                            stackml_layer.add_component(
                                Component(id_=component.name, type_=component.type.split('Component')[0]))

        for link in drawio.links:
            source = drawio.get_node(link.source)
            sink = drawio.get_node(link.source)
            stackml_link = Link(source=source.name,
                                sink=sink.name, layer=link.Layer)
            self.links.append(stackml_link)

    def sort_tiers(self):
        tiers_name = ['Thing', 'Edge', 'Network', 'Platform', 'Application']
        sorted_tiers = []
        for tier in tiers_name:
            tier_ = self.get_tier(tier)
            if tier_:
                sorted_tiers.append(tier_)

        self.tiers = sorted_tiers



#Copyright STACKEO - INRIA 2020
# pylint: disable=redefined-builtin,broad-except

from __future__ import print_function

import json
import os
import sys
import shutil
import tempfile
import zipfile
from io import BytesIO

import click
import requests

from jinja2 import Environment, FileSystemLoader
from jsonschema import ValidationError
from prettytable import PrettyTable
from yaml.scanner import ScannerError

from Stkml.Diagram.Map.RegionNotFound import RegionNotFound
from Stkml.Drawio import DrawIO
from Stkml.Stkml import Stkml, StkmlSyntaxicErrorList, ModelException
from Stkml.Diagram import  StkmlDiagram
from Stkml import MODULE_DIR, __version__, STACKIHUB_URL, STKML_PACKAGES
from Stkml.Stkml.PackageZip import PackageZip
from Stkml.Stkml.StkmlPackage import StkmlPackage

ROOT_DIR = os.path.dirname(os.path.abspath(__file__))


CONTEXT_SETTINGS = {'help_option_names':['-h', '--help']}
ICONS = os.path.join(MODULE_DIR, 'templates/diagram/resources/')
@click.group()
@click.version_option(__version__)
def cli():
    latest_version = get_stkml_latest_version()
    if latest_version:
        if __version__ != latest_version:
            click.secho(f"""[WARN] deprecated stkml@{__version__}: stkml@<{latest_version}\
 is no longer maintained. Upgrade to stkml@^{latest_version} using\
"pip install --upgrade stkml" """,
                        fg='yellow')

@click.command('check', help='check a Stkml project', context_settings=CONTEXT_SETTINGS)
@click.argument('stkmlfolder')
@click.option('--disable-syntax-verification', '-d', help='ignore syntax verification', is_flag=True)
def check(stkmlfolder: str, disable_syntax_verification: bool) -> None:
    stkml, _ = create_stkml(stkmlfolder)
    try:
        check_stkml(stkml, disable_syntax_verification)
    except click.ClickException as error:
        click.secho(f"[ERROR] {error}", fg='red')
        sys.exit(1)
    else:
        click.secho("[SUCCESS] Project checked successfully", fg='green')


@click.group('compile', help='compile a Stkml project using a specific backend',
             context_settings=CONTEXT_SETTINGS)
@click.argument('stkmlproject')
@click.option('--disable-syntax-verification', '-d', help='ignore syntax verification', is_flag=True)
@click.pass_context
def compile(ctx: click.Context, stkmlproject: str, disable_syntax_verification: bool) -> None:
    if stkmlproject:
        stkml, working_dir = create_stkml(stkmlproject)
        try:
            check_stkml(stkml, disable_syntax_verification)
        except click.ClickException as error:
            click.secho(f"[ERROR] {error}", fg='red')
            sys.exit(1)
        try:
            warns = stkml.create_architecture()
        except ModelException as error:
            click.secho(f"[ERROR] {error} ", fg='red')
            sys.exit(1)
        else:
            for warn in warns:
                click.secho(f"[WARN] File Not Found on  {warn} ", fg='yellow')
        ctx.ensure_object(dict)
        ctx.obj['Stkml'] = stkml
        ctx.obj['working_dir'] = working_dir

@click.command('diagram', short_help='compile a Stkml project for diagram', context_settings=CONTEXT_SETTINGS)
@click.option('--type', '-t', 'type', help='the diagram type {Architecture ,Topology or Map}',
              required=True, type=click.Choice(['1', '2', '3']))
@click.option('--output', '-o', 'outputfile', required=True, help='the output file')
@click.pass_context
def diagram(ctx: click.Context, type: str, outputfile: str) -> None:
    """
    compile Stkml file for diagram

    Example:
    \b
        Stkml compile stkml_project diagram -t 1 -o img
    """
    if which('dot'):
        stkml = ctx.obj['Stkml']
        working_dir = ctx.obj['working_dir']
        outputfile = os.path.join(working_dir, outputfile)
        diagram_ = StkmlDiagram()
        try:
            result = diagram_.diagram_from_stkml(type_=int(type),
                                                   stkml=stkml, output=outputfile)
        except FileNotFoundError as error:
            click.secho(f"[ERROR] {error} ", fg='red')
        except RegionNotFound as error:
            click.secho(f"[ERROR] {error} ", fg='red')
        else:
            if result:
                click.secho(result.rstrip().replace('Warning:', '[WARN]'), fg='yellow')
            click.secho("[SUCCESS] Compilation completed successfully", fg='green')
    else:
        click.secho("""[WARN] failed to execute 'dot',
                    make sure the Graphviz 'https://www.graphviz.org/' executables are on your systems' PATH""",
                    fg='yellow')


@click.command('drawio', short_help='compile a Stkml project for Drawio', context_settings=CONTEXT_SETTINGS)
@click.option('--level', '-l', 'level', help='the diagram view level {System View or Layer View}', required=True,
              type=click.Choice(['1', '2']))
@click.option('--icons', '-i', 'icons', default=ICONS, help='the icon folder {it contains [Node_Id].png files}')
@click.option('--output', '-o', 'outputfile', required=True, help='the output file')
@click.pass_context
def drawio(ctx: click.Context, level: str, icons: str, outputfile: str) -> None:
    """
    compile Stkml file for Drawio

    Example:
    \b
        Stkml compile stkml_project Drawio -l 1 -o Drawio
    """

    stkml = ctx.obj['Stkml']
    working_dir = ctx.obj['working_dir']
    outputfile = os.path.join(working_dir, outputfile)
    template_dir = os.path.join(MODULE_DIR, 'templates')
    env = Environment(loader=FileSystemLoader(template_dir))
    drawio_ = DrawIO(default_icons=ICONS, icons=icons)
    erreur = drawio_.from_stkml(stkml)
    if erreur:
        click.secho(f"[Error] {erreur} ", fg='red')
        sys.exit(1)
    warnings = drawio_.generate_drawio_diagram(output=outputfile, env=env, level=int(level))
    if len(warnings) > 0:
        for warn in warnings:
            click.secho(f"[WARN] {warn} ", fg='yellow')

    click.secho("[SUCCESS] Compilation completed successfully", fg='green')



@click.command('init', help='initialize a Stkml project', context_settings=CONTEXT_SETTINGS)
@click.argument('stkmldir')
def init(stkmldir: str) -> None:
    working_dir = os.path.realpath(os.getcwd())
    project_dir = check_folder_empty(working_dir, stkmldir)
    if project_dir:
        template_dir = os.path.join(MODULE_DIR, 'templates')
        stkml = Stkml()
        stkml.topology.name = click.prompt('Enter the system Name', type=str)
        stkml.topology.usecase = click.prompt('Enter the system use case', type=str)
        env = Environment(loader=FileSystemLoader(template_dir))
        output = stkml.generate_files(environment=env)
        outputfile = os.path.join(project_dir, 'main.stkml.yaml')
        with open(outputfile, 'w') as stkml_file:
            stkml_file.write(output)
        click.secho("[SUCCESS] Stkml project initialized successfully", fg='green')
    else:
        click.secho(f"[ERROR] can not create stkml project on {stkmldir}")


def check_stkml(stkml, disable_syntax_verification):
    try:
        stkml.check_stkml_project()
    except ScannerError as error:
        raise click.ClickException(f'YAML is not valid, {stkml.stkmlfile} {error}')
    except FileNotFoundError as error:
        raise click.ClickException(error.__str__())
    click.secho("[INFO] YAML is valid", fg='blue')
    if disable_syntax_verification:
        click.secho("[INFO] Ignore syntax verification", fg='blue')
    else:
        try:
            stkml.validate_stkml_project()
        except FileNotFoundError as error:
            raise click.ClickException(f'{error} not found')
        except StkmlSyntaxicErrorList as error_list:
            raise click.ClickException(f'Schema validation No\n {error_list}')
        except Exception as error:
            raise click.ClickException(error)
        click.secho("[INFO] Schema validation Ok", fg='blue')

def _check_package(stkml, disable_syntax_verification=False, check_metadata=False):
    try:
        stkml.check_stkml_package(check_metadata)
    except ScannerError as error:
        raise click.ClickException(f'YAML is not valid, {error}')
    except ValidationError as error:
        raise click.ClickException(f'The stkml metadata file is not valid, {error}')
    except FileNotFoundError as error:
        raise click.ClickException(f'{error}')
    click.secho("[INFO] YAML is valid", fg='blue')
    if disable_syntax_verification:
        click.secho("[INFO] Ignore syntax verification", fg='blue')
    else:
        try:
            result = stkml.validate_stkml_package()
        except FileNotFoundError as error:
            raise click.ClickException(f'{error} not found')
        except StkmlSyntaxicErrorList as error_list:
            raise click.ClickException(f'Schema validation No\n {error_list}')
        except Exception as error:
            raise click.ClickException(error)
        else:
            if result is True:
                click.secho("[INFO] Schema validation Ok", fg='blue')
            else:
                click.secho("[ERROR] Cant not find icon(s) :", fg='yellow')
                print_icon_not_found(result)
                sys.exit(1)


def print_icon_not_found(result):
    for key, values in result.items():
        click.secho(f"\t\t{key} : ", fg='yellow')
        for value in values:
            click.secho(f"\t\t\t{value}", fg='yellow')


def create_stkml(stkml_folder, ignore_main=False):
    working_dir = os.path.realpath(os.getcwd())
    stkml_folder = os.path.join(working_dir, stkml_folder)
    if not ignore_main:
        if not os.path.isdir(stkml_folder):
            raise click.BadParameter(message=f'{stkml_folder} doesn\'t exist',
                                     param=stkml_folder, param_hint='--input')
        if not os.path.isfile(os.path.join(stkml_folder, 'main.stkml.yaml')):
            raise click.BadParameter(message=f'{os.path.join(stkml_folder,"main.stkml.yaml") } doesn\'t exist',
                                     param=stkml_folder, param_hint='--input')
    stkml = Stkml(stkml_folder)
    return stkml, working_dir

def check_folder_empty(working_dir, outputdir):
    output_dir = os.path.join(working_dir, outputdir)
    if os.path.isdir(output_dir):
        if len(os.listdir(outputdir)) != 0:
            click.secho(f"[WARN] {output_dir} is not empty", fg='yellow')
            if click.confirm('Do you want to delete its content', abort=False):
                shutil.rmtree(output_dir)
                os.makedirs(output_dir)
                return output_dir
            return None
        return output_dir
    if click.confirm(f'{output_dir} Does not existe, do you want to create it', abort=False):
        os.makedirs(output_dir)
        return output_dir
    return None


def which(program):
    return shutil.which(program)

def get_stkml_latest_version() -> str or None:
    try:
        request = requests.get('https://pypi.org/pypi/stkml/json')
    except requests.exceptions.ConnectionError :
        pass
    else:
        version = None
        if request.status_code == 200:
            stkml_pypi = json.loads(request.content.decode("utf-8"))
            version = list(stkml_pypi['releases'].keys())[-1]
        request.close()
        return version

@click.command('check:package', help='check a Stkml package', context_settings=CONTEXT_SETTINGS)
@click.argument('package_path')
@click.option('--disable-syntax-verification', '-d', help='ignore syntax verification', is_flag=True)
@click.option('--check-metadata', '-m', help='check the package metadata', is_flag=True)
def check_package(package_path: str, disable_syntax_verification: bool, check_metadata: bool) -> None:
    stkml, _ = create_stkml(package_path, ignore_main=True)
    try:
        _check_package(stkml, disable_syntax_verification, check_metadata)
    except Exception as error:
        click.secho(f'[ERROR] {error.__str__()}', fg='red')
        sys.exit(1)
    click.secho("[SUCCESS] Package checked successfully", fg='green')

@click.command('install', help='install a Stkml package from stackhub'
               , context_settings=CONTEXT_SETTINGS)
@click.argument('name')
@click.option('--version', '-v', 'version', help='the version of the package',
              default='latest')
def install(name: str, version: str) -> None:
    packages_dir = create_stkml_packages_folder()
    url = f'{STACKIHUB_URL}files/{name}/{version}'
    try:
        zip = requests.get(url, stream=True)
    except requests.exceptions.ConnectionError as error:
        click.secho(f'[ERROR] can not connect to the {url} {error}', fg='red')
    else:
        if zip.status_code == 200:
            package_path = os.path.join(packages_dir, name)
            create_package_folder(package_path)
            file_bytes = BytesIO(zip.content)
            with PackageZip(file_bytes) as zip_package:
                zip_package.extract_all(package_path)
                click.secho(f'[SUCCESS] {name} has been installed successfully', fg='green')
        else:
            click.secho(f'[ERROR] {name} can not be installed : {zip.reason}', fg='red')

def create_package_folder(package_path):
    if os.path.isdir(package_path):
        shutil.rmtree(package_path)
    os.makedirs(package_path)


def create_stkml_packages_folder() -> str:
    try:
        packages_dir = StkmlPackage.get_packages_dir(os.getcwd())
    except RecursionError:
        packages_dir = os.path.join(os.getcwd(), STKML_PACKAGES)
        os.makedirs(packages_dir)
    return packages_dir





@click.command('search', help='search for a Stkml package on stackhub'
               , context_settings=CONTEXT_SETTINGS)
@click.argument('package')
def search(package):
    url = f'{STACKIHUB_URL}/search?package={package}'
    packages = requests.get(url)
    if packages.status_code == 200:
        packages = packages.json()
        nb_package = len(packages)
        if nb_package > 0:
            click.secho(f'[INFO] {nb_package} package(s) founded', fg='green')
            results_table = PrettyTable(list(packages[0].keys()))
            for pack in packages:
                results_table.add_row(list(pack.values()))
            print(results_table)
        else:
            click.secho(f'[WARN] There are no package similar to {package}', fg='yellow')
    else:
        click.secho(f'[ERROR] can not search for {package} : {packages.reason}')

@click.command('create:package', help='create a stkml package'
               , context_settings=CONTEXT_SETTINGS)
@click.argument('package')
def create_package(package: str):
    working_dir = os.path.realpath(os.getcwd())
    package_dir = check_folder_empty(working_dir, package)
    if package_dir:
        template_dir = os.path.join(MODULE_DIR, 'templates')
        package = StkmlPackage(package_dir)
        package.name = click.prompt('Enter the package''s Name', type=str, default=package.name)
        package.version = click.prompt('Enter the package''s version', type=str, default='1.0.0')
        package.author = click.prompt('Enter the package''s author', type=str, default='anonymous')
        env = Environment(loader=FileSystemLoader(template_dir))
        package.generate_metadata(env)
        click.secho(f"[SUCCESS] {package.name} has been initialized successfully", fg='green')
    else:
        click.secho(f"[ERROR] can not create stkml package on {package}")

@click.command('publish', help='publish a Stkml package to stackhub'
               , context_settings=CONTEXT_SETTINGS)
@click.argument('package')
def publish_package(package: str):
    working_dir = os.path.realpath(os.getcwd())
    package_dir = os.path.join(working_dir, os.path.realpath(package))
    if os.path.isdir(package_dir):
        stkml, _ = create_stkml(package, ignore_main=True)
        try:
            _check_package(stkml, check_metadata=True)
        except Exception as error:
            click.secho(f'[ERROR] {error.__str__()}', fg='red')
            click.secho(f'[ERROR] {package} can not be published because it is invalid', fg='red')
            sys.exit(1)
        click.secho("[INFO] Package checked successfully", fg='blue')
        with tempfile.TemporaryDirectory() as tmpdirname:
            package_zip_name = f'{tmpdirname}/{os.path.basename(package_dir)}.zip'
            with PackageZip(package_zip_name, 'w') as package_zip:
                package_zip.zip(package_dir)
                with open(package_zip_name, 'rb') as zif_file_pointer:
                    url = f'{STACKIHUB_URL}upload/'
                    req = requests.post(url, files={"file": zif_file_pointer})
                    if req.status_code == 200:
                        click.secho(f'[SUCCESS] {package} has been published successfully', fg='green')
                    else:
                        click.secho(f'[ERROR] {package} can not be published because : {req.reason}')
    else:
        click.secho(f"[ERROR] {package} is not package directory", fg='red')



cli.add_command(check)
cli.add_command(compile)
cli.add_command(init)
cli.add_command(check_package)
cli.add_command(install)
cli.add_command(search)
cli.add_command(create_package)
cli.add_command(publish_package)
compile.add_command(diagram)
compile.add_command(drawio)

if __name__ == '__main__':
    cli()



#Copyright STACKEO - INRIA 2020
#!/usr/bin/python
import os
MODULE_DIR = os.path.realpath(os.path.join(__file__, '..'))
STKML_URL_SCHEMA = 'https://stkml.stackeo.io/_static/stackml.schema.json'
STKML_PACKAGES = '.stkml_packages'
STKML_EXTENSION = '.stkml.yaml'
ICONS = os.path.join(MODULE_DIR, 'templates/diagram/resources/')
__version__ = '1.0.3'
STACKIHUB_URL = 'http://localhost/api/v1/'
STKML_METADATA = os.path.join(MODULE_DIR, 'templates', 'StkmlIn', 'stkml.yaml.json')
